/*
 * This software is in the public domain under CC0 1.0 Universal plus a
 * Grant of Patent License.
 * 
 * To the extent possible under law, the author(s) have dedicated all
 * copyright and related and neighboring rights to this software to the
 * public domain worldwide. This software is distributed without any
 * warranty.
 * 
 * You should have received a copy of the CC0 Public Domain Dedication
 * along with this software (see the LICENSE.md file). If not, see
 * <http://creativecommons.org/publicdomain/zero/1.0/>.
 */
package org.moqui.context;

import java.util.Map;

/** For localization (l10n) functionality, like localizing messages. */
public interface L10nFacade {

    /** Use the current locale (see ec.user.getLocale() method) to localize the message based on data in the
     * moqui.basic.LocalizedMessage entity. The localized message may have variables inserted using the ${} syntax that
     * when this is called through ec.resource.expand().
     *
     * The approach here is that original messages are actual messages in the primary language of the application. This
     * reduces issues with duplicated messages compared to the approach of explicit/artificial property keys. Longer
     * messages (over 255 characters) should use an artificial message key with the actual value always coming
     * from the database.
     */
    String localize(String original);

    /** Old method, still supported. See localize(). */
    @Deprecated
    String getLocalizedMessage(String original);

    /** Format currency amount for user to view.
     * @param amount An object representing the amount, should be a subclass of Number.
     * @param uomId The uomId (ISO currency code), required.
     * @param fractionDigits Number of digits after the decimal point to display. If null defaults to 2.
     * @return The formatted currency amount.
     */
    String formatCurrency(Object amount, String uomId, Integer fractionDigits);

    /** Format a Number, Timestamp, Date, Time, or Calendar object using the given format string. If no format string
     * is specified the default for the user's locale and time zone will be used.
     *
     * @param value The value to format. Must be a Number, Timestamp, Date, Time, or Calendar object.
     * @param format The format string used to specify how to format the value.
     * @return The value as a String formatted according to the format string.
     */
    String format(Object value, String format);
    /** Same as the format() method, exists to support code using this older method name. */
    @Deprecated
    String formatValue(Object value, String format);

    java.sql.Time parseTime(String input, String format);

    java.sql.Date parseDate(String input, String format);

    java.sql.Timestamp parseTimestamp(String input, String format);

    java.util.Calendar parseDateTime(String input, String format);

    java.math.BigDecimal parseNumber(String input, String format);
}
